#include <algorithm>
#include "skeleton2d.h"
#include "simplelexer.h"

using namespace std;


void HermiteSpline::
sort()
{
   std::sort(control.begin(), control.end());
}

void HermiteSpline::
compute_Catmull_Rom_slopes()
{
   // assume the control points have been sorted by knot value (increasing)
   // use formulas to set the slopes at each control point to the Catmull-Rom value
   // @@@@@@@@@@@@
}

float HermiteSpline::
operator()(float t)
{
   // do bisection search to find knot-interval in which t lies
   // then evaluate Hermite cubic and return value
   // @@@@@@@@@@@@
   return 0;
}

void Skeleton2D::
read_skeleton(istream &input)
{
   SimpleLexer lex(input);
   SimpleToken tok;

   // input format for simple skeleton data
   //   n        # number of links
   //            # and then for each link:
   //     i      # the number of the link (should be in order starting at 0)
   //     l_i    # the length of the link
   //     m_i    # the number of later links this link is attached to
   //            # and then for each attachment from the link:
   //       j    # the later link (j>i) this is attached to
   //       f    # the fraction along i at which it's attached

   lex.read(tok);
   if(tok.type!=TOKEN_INT) { cerr<<"Syntax of skeleton input file - expecting integer number of links"<<endl; exit(1); }
   link.resize(tok.data.int_value);

   for(unsigned int k=0; k<link.size(); ++k){
      lex.read(tok);
      if(tok.type!=TOKEN_INT) { cerr<<"Syntax of skeleton input file - expecting integer index for link #"<<k<<" in file"<<endl; exit(1); }
      int i=tok.data.int_value;
      if(i<0 || (unsigned int)i>=link.size()) { cerr<<"Illegal link number "<<i<<endl; exit(1); }

      lex.read(tok);
      if(tok.type==TOKEN_INT) link[i].length=(float)tok.data.int_value;
      else if(tok.type==TOKEN_FLOAT) link[i].length=tok.data.float_value;
      else { cerr<<"Syntax of skeleton input file - expecting number for length of link "<<i<<endl; exit(1); }

      lex.read(tok);
      if(tok.type!=TOKEN_INT) { cerr<<"Syntax of skeleton input file - expecting integer number of children for link "<<i<<endl; exit(1); }
      link[i].nchild=tok.data.int_value; 
      if(link[i].nchild<0) { cerr<<"Illegal number of children ("<<link[i].nchild<<") for link "<<i<<endl; exit(1); }

      for(int r=0; r<link[i].nchild; ++r){
         lex.read(tok);
         if(tok.type!=TOKEN_INT) { cerr<<"Syntax of skeleton input file - expecting integer index for attachment #"<<r<<" in link "<<i<<endl; exit(1); }
         int j=tok.data.int_value;
         if(j<=i || (unsigned int)j>=link.size()) { cerr<<"Illegal link number "<<j<<" in link "<<i<<endl; exit(1); }
         link[i].child.push_back(j);

         lex.read(tok);
         if(tok.type==TOKEN_INT) link[i].fraction.push_back((float)tok.data.int_value);
         else if(tok.type==TOKEN_FLOAT) link[i].fraction.push_back(tok.data.float_value);
         else { cerr<<"Syntax of skeleton input file - expecting number for fraction of link attachment "<<i<<"--"<<j<<endl; exit(1); }
      }
   }

   // make space for keyframe data
   joint.resize(link.size());
}

void Skeleton2D::
read_keyframes(istream &input)
{
   SimpleLexer lex(input);
   SimpleToken tok;

   // input format for keyframe data
   //   nframes              # number of frames to animate
   //                        # and then for each keyframe value either:
   //     t x xt             #   frame number, keyword "x", position xt of start endpoint of link 0
   //                        # or
   //     t y yt             #   frame number, keyword "y", position yt of start endpoint of link 0
   //                        # or
   //     t joint i theta    #   frame number, keyword "joint", link number, angle i should make from its parent
   //                            (or for link 0, angle it should make from the vertical)

   lex.read(tok);
   if(tok.type!=TOKEN_INT) { cerr<<"Syntax of keyframe input file - expecting integer number of frames"<<endl; exit(1); }
   nframes=tok.data.int_value;

   for(;;){
      lex.read(tok);
      if(tok.type==TOKEN_EOF) break;
      float t;
      if(tok.type==TOKEN_INT) t=(float)tok.data.int_value;
      else if(tok.type==TOKEN_FLOAT) t=tok.data.float_value;
      else { cerr<<"Syntax of keyframe input file - expecting number for time of key"<<endl; exit(1); }

      lex.read(tok);
      if(tok.type!=TOKEN_IDENTIFIER) { cerr<<"Syntax of keyframe input file - expecting keyword 'base' or 'joint' at time "<<t<<endl; exit(1); }
      if(strcmp("x", tok.data.identifier_string)==0){
         float x;

         lex.read(tok);
         if(tok.type==TOKEN_INT) x=(float)tok.data.int_value;
         else if(tok.type==TOKEN_FLOAT) x=tok.data.float_value;
         else { cerr<<"Syntax of skeleton input file - expecting number for x coordinate at time "<<t<<endl; exit(1); }

         link0x.control.push_back(ControlPoint(t, x));

      }else if(strcmp("y", tok.data.identifier_string)==0){
         float y;

         lex.read(tok);
         if(tok.type==TOKEN_INT) y=(float)tok.data.int_value;
         else if(tok.type==TOKEN_FLOAT) y=tok.data.float_value;
         else { cerr<<"Syntax of skeleton input file - expecting number for y coordinate at time "<<t<<endl; exit(1); }

         link0y.control.push_back(ControlPoint(t, y));

      }else if(strcmp("joint", tok.data.identifier_string)==0){
         int i;
         float theta;

         lex.read(tok);
         if(tok.type!=TOKEN_INT) { cerr<<"Syntax of keyframe input file - expecting integer link in joint at time "<<t<<endl; exit(1); }
         i=tok.data.int_value;

         lex.read(tok);
         if(tok.type==TOKEN_INT) theta=(float)tok.data.int_value;
         else if(tok.type==TOKEN_FLOAT) theta=tok.data.float_value;
         else { cerr<<"Syntax of skeleton input file - expecting number for theta coordinate of base at time "<<t<<endl; exit(1); }

         joint[i].control.push_back(ControlPoint(t, theta));

      }else{
         cerr<<"Syntax of keyframe input file - expecting keyword 'x', 'y', or 'joint' at time "<<t<<endl; exit(1);
      }
   }
}

void Skeleton2D::
compute_Catmull_Rom_slopes()
{
   link0x.sort();
   link0x.compute_Catmull_Rom_slopes();

   link0y.sort();
   link0y.compute_Catmull_Rom_slopes();

   for(unsigned int i=0; i<joint.size(); ++i){
      joint[i].sort();
      joint[i].compute_Catmull_Rom_slopes();
   }
}

void Skeleton2D::
write_animation(ostream &output)
{
   // output format for animation
   //   nframes              # number of frames in animation
   //                        # and then for each frame:
   //     n                  # number of links
   //                        # and then for each link:
   //       i                # the index of the link (doesn't have to be in order)
   //       x0 y0            # (x,y) position of start endpoint of link
   //       x1 y1            # (x,y) position of end endpoint of link

   output<<nframes<<endl;
   for(int frame=0; frame<nframes; ++frame){
      output<<"# frame "<<frame<<endl;
      output<<link.size()<<endl;
   
      // here's the forward kinematics code: traverse the skeleton, figuring out the endpoint positions of each link
      // @@@@@@@@@@@@
   }
}

