#include <iostream>
#include <fstream>
#include "simplelexer.h"

using namespace std;

struct Segment{
   float x0, y0, x1, y1;
};

struct Frame
{
   vector<Segment> segment;
};


void read_animation(const char *animfile, vector<Frame> &anim)
{
   ifstream in(animfile);
   SimpleLexer lex(in);
   SimpleToken tok;

   int nframes;
   lex.read(tok);
   if(tok.type!=TOKEN_INT) { cerr<<"Synatx of animation output file - expecting integer for number of frames"<<endl; exit(1); }
   nframes=tok.data.int_value;
   anim.resize(nframes);

   for(int i=0; i<nframes; ++i){
      int nlinks;
      lex.read(tok);
      if(tok.type!=TOKEN_INT) { cerr<<"Synatx of animation output file - expecting integer for number of links in frame "<<i<<endl; exit(1); }
      nlinks=tok.data.int_value;
      anim[i].segment.resize(nlinks);
      
      for(int j=0; j<nlinks; ++j){
	 int link;
         lex.read(tok);
         if(tok.type!=TOKEN_INT) { cerr<<"Synatx of animation output file - expecting integer for link number in frame "<<i<<endl; exit(1); }
	 link=tok.data.int_value;
	 if(link<0 || link>=nlinks) { cerr<<"Bad link number "<<link<<" in frame <<"<<i<<endl; exit(1); }

	 lex.read(tok);
	 if(tok.type==TOKEN_INT) anim[i].segment[link].x0=(float)tok.data.int_value;
	 else if(tok.type==TOKEN_FLOAT) anim[i].segment[link].x0=tok.data.float_value;
	 else { cerr<<"Syntax of animation output file - expecting number for x0 in link "<<link<<" of frame "<<i<<endl; exit(1); }

	 lex.read(tok);
	 if(tok.type==TOKEN_INT) anim[i].segment[link].y0=(float)tok.data.int_value;
	 else if(tok.type==TOKEN_FLOAT) anim[i].segment[link].y0=tok.data.float_value;
	 else { cerr<<"Syntax of animation output file - expecting number for y0 in link "<<link<<" of frame "<<i<<endl; exit(1); }

	 lex.read(tok);
	 if(tok.type==TOKEN_INT) anim[i].segment[link].x1=(float)tok.data.int_value;
	 else if(tok.type==TOKEN_FLOAT) anim[i].segment[link].x1=tok.data.float_value;
	 else { cerr<<"Syntax of animation output file - expecting number for x1 in link "<<link<<" of frame "<<i<<endl; exit(1); }

	 lex.read(tok);
	 if(tok.type==TOKEN_INT) anim[i].segment[link].y1=(float)tok.data.int_value;
	 else if(tok.type==TOKEN_FLOAT) anim[i].segment[link].y1=tok.data.float_value;
	 else { cerr<<"Syntax of animation output file - expecting number for y1 in link "<<link<<" of frame "<<i<<endl; exit(1); }
      }
   }
}

void read_camera(const char *camfile, char *&cam)
{
   ifstream in(camfile);
   vector<char> camvec;
   for(;;){
      char c=in.get();
      if(in.good()) camvec.push_back(c);
      else break;
   }
   cam=new char[camvec.size()+1];
   unsigned int i;
   for(i=0; i<camvec.size(); ++i)
      cam[i]=camvec[i];
   cam[i]=0;
}

void output_frame(const char *basename, int frame_number, const Frame &frame, const char *camera)
{
   // figure out the RIB file name, and open it
   char *filename=new char[strlen(basename)+20];
   sprintf(filename, "%s%04d.rib", basename, frame_number);
   ofstream out(filename);

   // first line: what image file this RIB file should produce
   sprintf(filename, "%s%04d.tiff", basename, frame_number);
   out<<"Display \""<<filename<<"\" \"file\" \"rgb\""<<endl;
   delete[] filename;

   // next line: image format (width and height in pixels, pixel aspect ratio)
   out<<"Format 640 480 1"<<endl;

   // sampling rate - 3x3 is pretty good
   out<<"PixelSamples 3 3"<<endl;

   // correct for a monitor with gamma 2.2 (we won't be compositing these images!)
   out<<"Exposure 1.0 2.2"<<endl;

   // then write out the camera specification
   out<<camera<<endl;

   // start the scene
   out<<"WorldBegin"<<endl;

   // lights
   out<<"LightSource \"ambientlight\" 1 \"intensity\" .4 \"lightcolor\" [1 1 1]"<<endl;
   out<<"LightSource \"distantlight\" 2 \"intensity\" 0.8 \"from\" [-5 30 20] \"to\" [0 0 0]"<<endl;
   out<<"LightSource \"distantlight\" 2 \"intensity\" 0.3 \"from\" [-5 -10 20] \"to\" [0 0 0]"<<endl;
   out<<"LightSource \"distantlight\" 2 \"intensity\" 0.3 \"from\" [5 -20 5] \"to\" [0 0 0]"<<endl;

   // set up material for stick figure
   out<<"Color [0.7 0.7 0.9]"<<endl;
   out<<"Surface \"plastic\" \"specularcolor\" [1 1 1]"<<endl;

   // output the geometry
   for(unsigned int i=0; i<frame.segment.size(); ++i){
      float x0=frame.segment[i].x0, y0=frame.segment[i].y0, x1=frame.segment[i].x1, y1=frame.segment[i].y1;
      // begin transformations to stretch and rotate sphere by pushing current state onto stack
      out<<"TransformBegin"<<endl;

      // scale and rotate to align with segment
      float r=0.12; // ratio of minor to major axis
      out<<"ConcatTransform ["<<0.5*(x0-x1)<<" "<<0.5*(y0-y1)<<" "<<" 0 0 "<<-r*(y1-y0)<<" "<<r*(x1-x0)<<" 0 0 0 0 "<<r<<" 0 "
	                      <<0.5*(x0+x1)<<" "<<0.5*(y0+y1)<<" 0 1]"<<endl;

      // sphere will become ellipsoid aligned with segment
      out<<"Sphere 1 -1 1 360"<<endl;

      //and we're done this part of the skeleton
      out<<"TransformEnd"<<endl; 
   }

   // finish the scene
   out<<"WorldEnd"<<endl;
}

int main(int argc, char **argv)
{
   if(argc!=4){
      cerr<<"Expecting three arguments: output_basename animation_file rib_camera_file"<<endl;
      return 1;
   }

   vector<Frame> animation;
   char *camera=0;
   read_animation(argv[2], animation);
   read_camera(argv[3], camera);
   for(unsigned int i=0; i<animation.size(); ++i){
      output_frame(argv[1], i, animation[i], camera);
   }
   delete[] camera;

   return 0;
}

